"""
Tools for working directly with OpenAI's various models.
"""
import logging
from typing import List, Union

from notdiamond import NotDiamond
from notdiamond.llms.providers import NDLLMProviders
from notdiamond.settings import NOTDIAMOND_API_KEY, OPENAI_API_KEY

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)

_ND_PARAMS = [
    "llm_configs",
    "default",
    "max_model_depth",
    "latency_tracking",
    "hash_content",
    "tradeoff",
    "preference_id",
    "tools",
    "callbacks",
    "nd_api_url",
    "nd_api_key",
    "user_agent",
]
_SHARED_PARAMS = ["timeout", "max_retries"]


class _OpenAIBase:
    """
    Base class which wraps both an openai client and Not Diamond retry / fallback logic.
    """

    def __init__(self, oai_client_cls, *args, **kwargs):
        nd_kwargs = {
            k: v for k, v in kwargs.items() if k in _ND_PARAMS + _SHARED_PARAMS
        }

        # TODO [a9] remove llm_configs as valid constructor arg for ND client
        self._nd_client = NotDiamond(
            api_key=nd_kwargs.get("nd_api_key", NOTDIAMOND_API_KEY),
            llm_configs=["openai/gpt-3.5-turbo"],
            *args,
            **nd_kwargs,
        )

        # Create a OpenAI client with a dummy model - will ignore this during routing
        oai_kwargs = {k: v for k, v in kwargs.items() if k not in _ND_PARAMS}
        self._oai_client = oai_client_cls(
            *args, api_key=OPENAI_API_KEY, **oai_kwargs
        )

    def __getattr__(self, name):
        return getattr(self._oai_client, name)

    def __call__(self, *args, **kwargs):
        return self._oai_client(*args, **kwargs)

    def __dir__(self):
        return dir(self._oai_client)

    @property
    def chat(self):
        class ChatCompletions:
            def __init__(self, parent):
                self.parent = parent

            @property
            def completions(self):
                return self

            def create(self, *args, **kwargs):
                return self.parent.create(*args, **kwargs)

        return ChatCompletions(self)

    def _create_prep(self, model: Union[str, List], **kwargs):
        model = kwargs.get("model", model)

        if model is None:
            LOGGER.info(
                "No LLM configs provided. Not Diamond will route to all OpenAI models."
            )
            llm_configs = [
                str(p) for p in NDLLMProviders if p.provider == "openai"
            ]
        elif isinstance(model, str):
            llm_configs = model.split(",")
        elif isinstance(model, list):
            llm_configs = self._nd_client._parse_llm_configs_data(model)

        if "messages" not in kwargs:
            raise ValueError("'messages' argument is required")

        return llm_configs


class OpenAI(_OpenAIBase):
    """
    Encapsulating class for an openai.OpenAI client. This supports the same methods as
    the openai package, while also supporting routed prompts with calls to `completion`.
    """

    def __init__(self, *args, **kwargs):
        from openai import OpenAI as OpenAIClient

        super().__init__(OpenAIClient, *args, **kwargs)

    def create(self, *args, model: Union[str, List] = None, **kwargs):
        """
        Perform chat completion using OpenAI's API, after routing the prompt to a
        specific LLM via Not Diamond.
        """
        llm_configs = self._create_prep(model, **kwargs)
        session_id, best_llm = self._nd_client.model_select(
            *args, model=llm_configs, **kwargs
        )
        response = self._oai_client.chat.completions.create(
            *args, model=str(best_llm.model), **kwargs
        )
        LOGGER.info(f"Routed prompt to {best_llm} for session ID {session_id}")
        return response


class AsyncOpenAI(_OpenAIBase):
    """
    Encapsulating class for an openai.OpenAI client. This supports the same methods as
    the openai package, while also supporting routed prompts with calls to `completion`.
    """

    def __init__(self, *args, **kwargs):
        from openai import AsyncOpenAI as OpenAIClient

        super().__init__(OpenAIClient, *args, **kwargs)

    async def create(self, *args, model: Union[str, List] = None, **kwargs):
        """
        Perform async chat completion using OpenAI's API, after routing the prompt to a
        specific LLM via Not Diamond.
        """
        llm_configs = self._create_prep(model, **kwargs)
        session_id, best_llm = await self._nd_client.amodel_select(
            *args, model=llm_configs, **kwargs
        )
        response = await self._oai_client.chat.completions.create(
            *args, model=str(best_llm.model), **kwargs
        )
        LOGGER.debug(
            f"Routed prompt to {best_llm} for session ID {session_id}"
        )
        return response



from dataclasses import dataclass
from typing import Any, ClassVar, Dict, List, Optional, Type, Union, get_args

import optuna

from notdiamond.toolkit.rag.evaluation_dataset import RAGEvaluationDataset


@dataclass
class IntValueRange:
    """
    A range of int values for an auto-evaluated RAG pipeline. Useful for, eg. RAG context chunk size.
    """

    lo: int
    hi: int
    step: int

    def __contains__(self, value: int) -> bool:
        return (
            self.lo <= value <= self.hi and (value - self.lo) % self.step == 0
        )


@dataclass
class FloatValueRange:
    """
    A range of float values for an auto-evaluated RAG pipeline. Useful for, eg. LLM temperature.
    """

    lo: float
    hi: float
    step: float

    def __contains__(self, value: float) -> bool:
        return self.lo <= value <= self.hi


@dataclass
class CategoricalValueOptions:
    """
    A list of categorical values for an auto-evaluated RAG pipeline. Useful for, eg. embedding algorithms.
    """

    values: List[str]

    def __contains__(self, value: str) -> bool:
        return value in self.values


_ALLOWED_TYPES = [IntValueRange, FloatValueRange, CategoricalValueOptions]


class BaseNDRagWorkflow:
    """
    A base interface for a RAG workflow to be auto-evaluated by Not Diamond.

    Subclasses should define parameter_specs to type parameters they need to optimize,
    by using type annotations with the above dataclasses. For example:

    .. code-block:: python

        class ExampleNDRagWorkflow(BaseNDRagWorkflow):
            parameter_specs = {
                "chunk_size": (Annotated[int, IntValueRange(1000, 2500, 500)], 1000),
                                "chunk_overlap": (Annotated[int, IntValueRange(50, 200, 25)], 100),
                                "top_k": (Annotated[int, IntValueRange(1, 20, 1)], 5),
                                "algo": (
                                        Annotated[
                                                str,
                                                CategoricalValueOptions(
                                                        [
                                                                "BM25",
                                                                "openai_small",
                                                                "openai_large",
                                                                "cohere_eng",
                                                                "cohere_multi",
                                                        ]
                        ),
                    ],
                    "BM25",
                ),
                "temperature": (Annotated[float, FloatValueRange(0.0, 1.0, 0.1)], 0.9),
            }

    """

    parameter_specs: ClassVar[Dict[str, tuple[Type, Any]]] = {}

    def __init__(
        self,
        documents: Any,
        test_queries: Optional[Any] = None,
        objective_maximize: bool = True,
        **kwargs,
    ):
        """
        Args:
            evaluation_dataset: A dataset of RAG evaluation samples.
            documents: The documents to use for RAG.
            objective_maximize: Whether to maximize or minimize the objective defined below.
        """
        if not self.parameter_specs:
            raise NotImplementedError(
                f"Class {self.__class__.__name__} must define parameter_specs"
            )

        self._param_ranges = {}
        self._base_param_types = {}
        for param_name, (
            param_type,
            default_value,
        ) in self.parameter_specs.items():
            type_args = get_args(param_type)
            base_type, range_type = type_args
            if range_type is None:
                raise ValueError(
                    f"Expected parameter type in {_ALLOWED_TYPES} but received {param_type}"
                )
            self._param_ranges[param_name] = range_type
            self._base_param_types[param_name] = base_type
            if not isinstance(default_value, base_type):
                raise ValueError(
                    f"Expected default value type {base_type} but received {type(default_value)}"
                )
            setattr(self, param_name, default_value)

        self.documents = documents
        self.test_queries = test_queries
        self.objective_maximize = objective_maximize

    def get_parameter_type(self, param_name: str) -> Type:
        param_type = self._base_param_types.get(param_name)
        if param_type is None:
            raise ValueError(
                f"Parameter {param_name} not found in parameter_specs"
            )
        return param_type

    def get_parameter_range(self, param_name: str) -> Type:
        param_range = self._param_ranges.get(param_name)
        if param_range is None:
            raise ValueError(
                f"Parameter {param_name} not found in parameter_specs"
            )
        return param_range

    def rag_workflow(
        self, documents: Any, test_queries: Optional[List[str]]
    ) -> RAGEvaluationDataset:
        """
        Define RAG workflow components here by setting instance attrs on `self`. Those attributes will be set
        at init-time and available when retrieving context or generating responses.
        """
        raise NotImplementedError()

    def objective(self):
        """
        Define the objective function for your RAG workflow. The workflow's hyperparameters will be optimized
        according to values of this objective.
        """
        raise NotImplementedError()

    def _outer_objective(self, trial: optuna.Trial):
        for param_name in self.parameter_specs.keys():
            param_range = self.get_parameter_range(param_name)
            if isinstance(param_range, IntValueRange):
                param_value = trial.suggest_int(
                    param_name,
                    param_range.lo,
                    param_range.hi,
                    step=param_range.step,
                )
            elif isinstance(param_range, FloatValueRange):
                param_value = trial.suggest_float(
                    param_name,
                    param_range.lo,
                    param_range.hi,
                    step=param_range.step,
                )
            elif isinstance(param_range, CategoricalValueOptions):
                param_value = trial.suggest_categorical(
                    param_name, param_range.values
                )
            else:
                raise ValueError(
                    f"Expected parameter type in {_ALLOWED_TYPES} but received unknown parameter type: {param_range}"
                )
            setattr(self, param_name, param_value)

        self.evaluation_dataset = self.rag_workflow(
            self.documents, self.test_queries
        )
        result = self.objective()
        self._reset_param_values()
        return result

    def _get_default_param_values(self):
        return {
            param_name: default_value
            for param_name, (_, default_value) in self.parameter_specs.items()
        }

    def _set_param_values(
        self, param_values: Dict[str, Union[int, float, str]]
    ):
        for param_name in self.parameter_specs.keys():
            param_value = param_values.get(param_name)
            param_type = self.get_parameter_type(param_name)
            param_range = self.get_parameter_range(param_name)
            if param_value is None:
                raise ValueError(
                    f"Best value for {param_name} not found. This should not happen."
                )
            elif not isinstance(param_value, param_type):
                raise ValueError(
                    f"Expected parameter type {param_type} but received {type(param_value)}"
                )
            elif param_value not in param_range:
                raise ValueError(
                    f"Parameter value {param_value} not in range {param_range}"
                )
            setattr(self, param_name, param_value)

    def _reset_param_values(self):
        for param_name in self.parameter_specs.keys():
            setattr(
                self, param_name, self._get_default_param_values()[param_name]
            )

    def job_name(self):
        raise NotImplementedError()

    def get_retrieved_context(self, query: str) -> List[str]:
        raise NotImplementedError()

    def get_response(self, query: str) -> str:
        raise NotImplementedError()



class UnsupportedLLMProvider(Exception):
    """The exception class for unsupported LLM provider"""


class UnsupportedEmbeddingProvider(Exception):
    """The exception class for unsupported Embedding provider"""


class InvalidApiKey(Exception):
    """The exception class for InvalidApiKey"""


class MissingApiKey(Exception):
    """The exception class for MissingApiKey"""


class MissingLLMConfigs(Exception):
    """The exception class for empty LLM provider configs array"""


class ApiError(Exception):
    """The exception class for any ApiError"""


class CreateUnavailableError(Exception):
    """The exception class raised when `notdiamond[create]` is not available"""



import logging
from typing import Optional

from notdiamond import settings
from notdiamond.exceptions import (
    UnsupportedEmbeddingProvider,
    UnsupportedLLMProvider,
)

POSSIBLE_PROVIDERS = list(settings.PROVIDERS.keys())
POSSIBLE_MODELS = list(
    model
    for provider_values in settings.PROVIDERS.values()
    for values in provider_values.values()
    if isinstance(values, list)
    for model in values
)

POSSIBLE_EMBEDDING_PROVIDERS = [
    *list(settings.EMBEDDING_PROVIDERS.keys()),
    "huggingface",
]
POSSIBLE_EMBEDDING_MODELS = list(
    model
    for provider_values in settings.EMBEDDING_PROVIDERS.values()
    for values in provider_values.values()
    if isinstance(values, list)
    for model in values
)

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


class LLMConfig:
    """
    A NotDiamond LLM provider config (or LLMConfig) is represented by a combination of provider and model.
    Provider refers to the company of the foundational model, such as openai, anthropic, google.
    The model represents the model name as defined by the owner company, such as gpt-3.5-turbo
    Beside this you can also specify the API key for each provider, specify extra arguments
    that are also supported by Langchain (eg. temperature), and a system prmopt to be used
    with the provider. If the provider is selected during routing, then the system prompt will
    be used, replacing the one in the message array if there are any.

    All supported providers and models can be found in our docs.

    If the API key it's not specified, it will try to pick it up from an .env file before failing.
    As example for OpenAI it will look for OPENAI_API_KEY.

    Attributes:
        provider (str): The name of the LLM provider (e.g., "openai", "anthropic"). Must be one of the
                        predefined providers in `POSSIBLE_PROVIDERS`.
        model (str): The name of the LLM model to use (e.g., "gpt-3.5-turbo").
                        Must be one of the predefined models in `POSSIBLE_MODELS`.
        system_prompt (Optional[str], optional): The system prompt to use for the provider. Defaults to None.
        api_key (Optional[str], optional): The API key for accessing the LLM provider's services.
                                            Defaults to None, in which case it tries to fetch from the settings.
        openrouter_model (str): The OpenRouter model equivalent for this provider / model
        **kwargs: Additional keyword arguments that might be necessary for specific providers or models.

    Raises:
        UnsupportedLLMProvider: If the `provider` or `model` specified is not supported.
    """

    def __init__(
        self,
        provider: str,
        model: str,
        is_custom: bool = False,
        system_prompt: Optional[str] = None,
        context_length: Optional[int] = None,
        input_price: Optional[float] = None,
        custom_input_price: Optional[float] = None,
        output_price: Optional[float] = None,
        custom_output_price: Optional[float] = None,
        latency: Optional[float] = None,
        custom_latency: Optional[float] = None,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        """_summary_

        Args:
            provider (str): The name of the LLM provider (e.g., "openai", "anthropic").
            model (str): The name of the LLM model to use (e.g., "gpt-3.5-turbo").
            is_custom (bool): Whether this is a custom model. Defaults to False.
            system_prompt (Optional[str], optional): The system prompt to use for the provider. Defaults to None.
            context_length (Optional[int], optional): Custom context window length for the provider/model.
            custom_input_price (Optional[float], optional): Custom input price (USD) per million tokens for this
                provider/model; will default to public input price if available.
            custom_output_price (Optional[float], optional): Custom output price (USD) per million tokens for this
                provider/model; will default to public output price if available.
            custom_latency (Optional[float], optional): Custom latency (time to first token) for provider/model.
            api_key (Optional[str], optional): The API key for accessing the LLM provider's services.
                                                Defaults to None.
            **kwargs: Additional keyword arguments that might be necessary for specific providers or models.

        Raises:
            UnsupportedLLMProvider: If the `provider` or `model` specified is not supported.
        """
        if is_custom:
            self._openrouter_model = None
            self.api_key = api_key
            self.default_input_price = custom_input_price or input_price
            self.default_output_price = custom_output_price or output_price
        else:
            if provider not in POSSIBLE_PROVIDERS:
                raise UnsupportedLLMProvider(
                    f"Given LLM provider {provider} is not in the list of supported providers."
                )
            if model not in POSSIBLE_MODELS:
                raise UnsupportedLLMProvider(
                    f"Given LLM model {model} is not in the list of supported models."
                )
            self._openrouter_model = settings.PROVIDERS[provider][
                "openrouter_identifier"
            ].get(model, None)

            self.api_key = (
                api_key
                if api_key is not None
                else settings.PROVIDERS[provider]["api_key"]
            )

            self.default_input_price = settings.PROVIDERS[provider]["price"][
                model
            ]["input"]
            self.default_output_price = settings.PROVIDERS[provider]["price"][
                model
            ]["output"]

        self.provider = provider
        self.model = model
        self.system_prompt = system_prompt

        self.is_custom = is_custom
        self.context_length = context_length
        self.input_price = custom_input_price or input_price
        self.output_price = custom_output_price or output_price
        self.latency = custom_latency or latency

        self.kwargs = kwargs

    def __str__(self) -> str:
        return f"{self.provider}/{self.model}"

    def __repr__(self) -> str:
        return f"LLMConfig({self.provider}/{self.model})"

    def __eq__(self, other):
        if isinstance(other, LLMConfig):
            return (
                self.provider == other.provider and self.model == other.model
            )
        return False

    def __hash__(self):
        return hash(str(self))

    @property
    def openrouter_model(self):
        if self._openrouter_model is None:
            LOGGER.warning(
                f"Configured model {str(self)} is not available via OpenRouter. Please try another model."
            )
        return self._openrouter_model

    def prepare_for_request(self):
        """
        Converts the LLMConfig object to a dict in the format accepted by
        the NotDiamond API.

        Returns:
            dict
        """
        return {
            "provider": self.provider,
            "model": self.model,
            "is_custom": self.is_custom,
            "context_length": self.context_length,
            "input_price": self.input_price,
            "output_price": self.output_price,
            "latency": self.latency,
        }

    def set_api_key(self, api_key: str) -> "LLMConfig":
        self.api_key = api_key

        return self

    @classmethod
    def from_string(cls, llm_provider: str):
        """
        We allow our users to specify LLM providers for NotDiamond in the string format 'provider_name/model_name',
        as example 'openai/gpt-3.5-turbo'. Underlying our workflows we want to ensure we use LLMConfig as
        the base type, so this class method converts a string specification of an LLM provider into an
        LLMConfig object.

        Args:
            llm_provider (str): this is the string definition of the LLM provider

        Returns:
            LLMConfig: initialized object with correct provider and model
        """
        split_items = llm_provider.split("/")
        if len(split_items) not in [2, 3]:
            raise ValueError(
                f"Expected string of format 'provider/model' or 'prefix/provider/model' but got {llm_provider}"
            )
        elif len(split_items) == 3:
            _, provider, model = split_items
        else:
            provider = split_items[0]
            model = split_items[1]
        return cls(provider=provider, model=model)


class EmbeddingConfig:
    """
    A NotDiamond embedding provider config (or EmbeddingConfig) is represented by a combination of provider and model.
    Provider refers to the company of the foundational model, such as openai, anthropic, google.
    The model represents the model name as defined by the owner company, such as text-embedding-3-large
    Beside this you can also specify the API key for each provider or extra arguments
    that are also supported by Langchain.

    All supported providers and models can be found in our docs.

    If the API key is not specified, the Config will try to read the key from an .env file before failing.
    For example, the Config will look for `OPENAI_API_KEY` to authenticate any OpenAI provider.

    Attributes:
        provider (str): The name of the LLM provider (e.g., "openai", "anthropic"). Must be one of the
                        predefined providers in `POSSIBLE_EMBEDDING_PROVIDERS`.
        model (str): The name of the LLM model to use (e.g., "gpt-3.5-turbo").
                        Must be one of the predefined models in `POSSIBLE_MODELS`.
        api_key (Optional[str], optional): The API key for accessing the LLM provider's services.
                                            Defaults to None, in which case it tries to fetch from the environment.
        **kwargs: Additional keyword arguments that might be necessary for specific providers or models.

    Raises:
        UnsupportedLLMProvider: If the `provider` or `model` specified is not supported.
    """

    def __init__(
        self,
        provider: str,
        model: str,
        api_key: Optional[str] = None,
        **kwargs,
    ):
        """_summary_

        Args:
            provider (str): The name of the embedding provider (e.g., "openai", "anthropic").
            model (str): The name of the embedding model to use (e.g., "text-embedding-3-large").
            api_key (Optional[str], optional): The API key for accessing the embedding provider's services.
                                                Defaults to None.
            **kwargs: Additional keyword arguments that might be necessary for specific providers or models.

        Raises:
            UnsupportedEmbeddingProvider: If the `provider` or `model` specified is not supported.
        """
        if provider not in POSSIBLE_EMBEDDING_PROVIDERS:
            raise UnsupportedEmbeddingProvider(
                f"Given embedding provider {provider} is not in the list of supported providers."
            )

        if (
            model not in POSSIBLE_EMBEDDING_MODELS
            and provider != "huggingface"
        ):
            raise UnsupportedEmbeddingProvider(
                f"Given embedding model {model} is not in the list of supported models."
            )

        self.api_key = (
            api_key
            if api_key is not None
            else settings.PROVIDERS[provider]["api_key"]
        )

        self.provider = provider
        self.model = model
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f"{self.provider}/{self.model}"

    def __repr__(self) -> str:
        return f"EmbeddingConfig({self.provider}/{self.model})"

    def __eq__(self, other):
        if isinstance(other, EmbeddingConfig):
            return (
                self.provider == other.provider and self.model == other.model
            )
        return False

    def __hash__(self):
        return hash(str(self))

    def set_api_key(self, api_key: str) -> "EmbeddingConfig":
        self.api_key = api_key

        return self

    @classmethod
    def from_string(cls, llm_provider: str):
        """
        We allow our users to specify LLM providers for NotDiamond in the string format 'provider_name/model_name',
        for example 'openai/gpt-3.5-turbo'. Our workflows expect LLMConfig as
        the base type, so this class method converts a string specification of an LLM provider into an
        LLMConfig object.

        Args:
            llm_provider (str): this is the string definition of the LLM provider

        Returns:
            LLMConfig: initialized object with correct provider and model
        """
        split_items = llm_provider.split("/")
        if len(split_items) not in [2, 3]:
            raise ValueError(
                f"Expected string of format 'provider/model' or 'prefix/provider/model' but got {llm_provider}"
            )
        elif len(split_items) == 3:
            _, provider, model = split_items
        else:
            provider = split_items[0]
            model = split_items[1]
        return cls(provider=provider, model=model)



import logging
from typing import Dict

import requests

from notdiamond import settings
from notdiamond._utils import _default_headers
from notdiamond.exceptions import ApiError
from notdiamond.llms.config import LLMConfig
from notdiamond.types import FeedbackRequestPayload

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def feedback_request(
    session_id: str,
    llm_config: LLMConfig,
    feedback_payload: Dict[str, int],
    notdiamond_api_key: str,
    nd_api_url: str = settings.NOTDIAMOND_API_URL,
    _user_agent: str = settings.DEFAULT_USER_AGENT,
) -> bool:
    url = f"{nd_api_url}/v2/report/metrics/feedback"

    payload: FeedbackRequestPayload = {
        "session_id": session_id,
        "provider": llm_config.prepare_for_request(),
        "feedback": feedback_payload,
    }

    headers = _default_headers(notdiamond_api_key, _user_agent)

    try:
        response = requests.post(url, json=payload, headers=headers)
    except Exception as e:
        raise ApiError(f"ND API error for feedback: {e}")

    if response.status_code != 200:
        LOGGER.error(
            f"ND API feedback error: failed to report feedback with status {response.status_code}. {response.text}"
        )
        return False

    return True



from typing import Any, Dict, List

from pydantic import BaseModel, field_validator

from notdiamond.exceptions import InvalidApiKey, MissingApiKey


class NDApiKeyValidator(BaseModel):
    api_key: str

    @field_validator("api_key", mode="before")
    @classmethod
    def api_key_must_be_a_string(cls, v) -> str:
        if not isinstance(v, str):
            raise InvalidApiKey("ND API key should be a string")
        return v

    @field_validator("api_key", mode="after")
    @classmethod
    def string_must_not_be_empty(cls, v):
        if len(v) == 0:
            raise MissingApiKey("ND API key should be longer than 0")
        return v


class ModelSelectRequestPayload(BaseModel):
    prompt_template: str
    formatted_prompt: str
    components: Dict[str, Dict]
    llm_configs: List[Dict]
    metric: str
    max_model_depth: int


class FeedbackRequestPayload(BaseModel):
    session_id: str
    provider: Dict[str, Any]
    feedback: Dict[str, int]



from enum import Enum

from notdiamond.llms.config import LLMConfig


class NDLLMProviders(Enum):
    """
    NDLLMProviders serves as a registry for the supported LLM models by NotDiamond.
    It allows developers to easily specify available LLM providers for the router.

    Attributes:
        GPT_3_5_TURBO (NDLLMProvider): refers to 'gpt-3.5-turbo' model by OpenAI
        GPT_3_5_TURBO_0125 (NDLLMProvider): refers to 'gpt-3.5-turbo-0125' model by OpenAI
        GPT_4 (NDLLMProvider): refers to 'gpt-4' model by OpenAI
        GPT_4_0613 (NDLLMProvider): refers to 'gpt-4-0613' model by OpenAI
        GPT_4_1106_PREVIEW (NDLLMProvider): refers to 'gpt-4-1106-preview' model by OpenAI
        GPT_4_TURBO (NDLLMProvider): refers to 'gpt-4-turbo' model by OpenAI
        GPT_4_TURBO_PREVIEW (NDLLMProvider): refers to 'gpt-4-turbo-preview' model by OpenAI
        GPT_4_TURBO_2024_04_09 (NDLLMProvider): refers to 'gpt-4-turbo-2024-04-09' model by OpenAI
        GPT_4o_2024_05_13 (NDLLMProvider): refers to 'gpt-4o-2024-05-13' model by OpenAI
        GPT_4o_2024_08_06 (NDLLMProvider): refers to 'gpt-4o-2024-08-06' model by OpenAI
        GPT_4o (NDLLMProvider): refers to 'gpt-4o' model by OpenAI
        GPT_4o_MINI_2024_07_18 (NDLLMProvider): refers to 'gpt-4o-mini-2024-07-18' model by OpenAI
        GPT_4o_MINI (NDLLMProvider): refers to 'gpt-4o-mini' model by OpenAI
        GPT_4_0125_PREVIEW (NDLLMProvider): refers to 'gpt-4-0125-preview' model by OpenAI
        GPT_4_1 (NDLLMProvider): refers to 'gpt-4.1' model by OpenAI
        GPT_4_1_2025_04_14 (NDLLMProvider): refers to 'gpt-4.1-2025-04-14' model by OpenAI
        GPT_4_1_MINI (NDLLMProvider): refers to 'gpt-4.1-mini' model by OpenAI
        GPT_4_1_MINI_2025_04_14 (NDLLMProvider): refers to 'gpt-4.1-mini-2025-04-14' model by OpenAI
        GPT_4_1_NANO (NDLLMProvider): refers to 'gpt-4.1-nano' model by OpenAI
        GPT_4_1_NANO_2025_04_14 (NDLLMProvider): refers to 'gpt-4.1-nano-2025-04-14' model by OpenAI
        GPT_4o_2024_11_20 (NDLLMProvider): refers to 'gpt-4o-2024-11-20' model by OpenAI

        CLAUDE_2_1 (NDLLMProvider): refers to 'claude-2.1' model by Anthropic
        CLAUDE_3_OPUS_20240229 (NDLLMProvider): refers to 'claude-3-opus-20240229' model by Anthropic
        CLAUDE_3_SONNET_20240229 (NDLLMProvider): refers to 'claude-3-sonnet-20240229' model by Anthropic
        CLAUDE_3_5_SONNET_20240620 (NDLLMProvider): refers to 'claude-3-5-sonnet-20240620' model by Anthropic
        CLAUDE_3_7_SONNET_LATEST (NDLLMProvider): refers to 'claude-3-7-sonnet-latest' model by Anthropic
        CLAUDE_3_7_SONNET_20250219 (NDLLMProvider): refers to 'claude-3-7-sonnet-20250219' model by Anthropic
        CLAUDE_3_5_HAIKU_20241022 (NDLLMProvider): refers to 'claude-3-5-haiku-20241022' model by Anthropic
        CLAUDE_3_HAIKU_20240307 (NDLLMProvider): refers to 'claude-3-haiku-20240307' model by Anthropic
        CLAUDE_OPUS_4_20250514 (NDLLMProvider): refers to 'claude-opus-4-20250514' model by Anthropic
        CLAUDE_OPUS_4_1_20250805 (NDLLMProvider): refers to 'claude-opus-4-1-20250805' model by Anthropic
        CLAUDE_SONNET_4_20250514 (NDLLMProvider): refers to 'claude-sonnet-4-20250514' model by Anthropic
        CLAUDE_SONNET_4_5 (NDLLMProvider): refers to 'claude-sonnet-4-5' model by Anthropic
        CLAUDE_SONNET_4_5_20250929 (NDLLMProvider): refers to 'claude-sonnet-4-5-20250929' model by Anthropic
        CLAUDE_OPUS_4_0 (NDLLMProvider): refers to 'claude-opus-4-0' model by Anthropic
        CLAUDE_OPUS_4_1 (NDLLMProvider): refers to 'claude-opus-4-1' model by Anthropic
        CLAUDE_SONNET_4_0 (NDLLMProvider): refers to 'claude-sonnet-4-0' model by Anthropic

        GEMINI_PRO (NDLLMProvider): refers to 'gemini-pro' model by Google
        GEMINI_1_PRO_LATEST (NDLLMProvider): refers to 'gemini-1.0-pro-latest' model by Google
        GEMINI_15_PRO_LATEST (NDLLMProvider): refers to 'gemini-1.5-pro-latest' model by Google
        GEMINI_15_PRO_EXP_0801 (NDLLMProvider): refers to 'gemini-1.5-pro-exp-0801' model by Google
        GEMINI_15_FLASH_LATEST (NDLLMProvider): refers to 'gemini-1.5-flash-latest' model by Google
        GEMINI_20_FLASH (NDLLMProvider): refers to 'gemini-20-flash' model by Google
        GEMINI_20_FLASH_001 (NDLLMProvider): refers to 'gemini-20-flash-001' model by Google
        GEMINI_25_FLASH (NDLLMProvider): refers to 'gemini-25-flash' model by Google
        GEMINI_25_PRO (NDLLMProvider): refers to 'gemini-25-pro' model by Google

        COMMAND_R (NDLLMProvider): refers to 'command-r' model by Cohere
        COMMAND_R_PLUS (NDLLMProvider): refers to 'command-r-plus' model by Cohere

        MISTRAL_LARGE_LATEST (NDLLMProvider): refers to 'mistral-large-latest' model by Mistral AI
        MISTRAL_LARGE_2407 (NDLLMProvider): refers to 'mistral-large-2407' model by Mistral AI
        MISTRAL_LARGE_2402 (NDLLMProvider): refers to 'mistral-large-2402' model by Mistral AI
        MISTRAL_MEDIUM_LATEST (NDLLMProvider): refers to 'mistral-medium-latest' model by Mistral AI
        MISTRAL_SMALL_LATEST (NDLLMProvider): refers to 'mistral-small-latest' model by Mistral AI
        OPEN_MISTRAL_7B (NDLLMProvider): refers to 'open-mistral-7b' model by Mistral AI
        OPEN_MIXTRAL_8X7B (NDLLMProvider): refers to 'open-mixtral-8x7b' model by Mistral AI
        OPEN_MIXTRAL_8X22B (NDLLMProvider): refers to 'open-mixtral-8x22b' model by Mistral AI
        OPEN_MISTRAL_NEMO (NDLLMProvider): refers to 'open-mistral-nemo' model by Mistral AI

        TOGETHER_MISTRAL_7B_INSTRUCT_V0_2 (NDLLMProvider): refers to 'Mistral-7B-Instruct-v0.2' model served via TogetherAI
        TOGETHER_MIXTRAL_8X7B_INSTRUCT_V0_1 (NDLLMProvider): refers to 'Mixtral-8x7B-Instruct-v0.1' model served via TogetherAI
        TOGETHER_MIXTRAL_8X22B_INSTRUCT_V0_1 (NDLLMProvider): refers to 'Mixtral-8x22B-Instruct-v0.1' model served via TogetherAI
        TOGETHER_LLAMA_3_70B_CHAT_HF (NDLLMProvider): refers to 'Llama-3-70b-chat-hf' model served via TogetherAI
        TOGETHER_LLAMA_3_8B_CHAT_HF (NDLLMProvider): refers to 'Llama-3-8b-chat-hf' model served via TogetherAI
        TOGETHER_QWEN2_72B_INSTRUCT (NDLLMProvider): refers to 'Qwen2-72B-Instruct' model served via TogetherAI
        TOGETHER_LLAMA_3_1_8B_INSTRUCT_TURBO (NDLLMProvider): refers to 'Meta-Llama-3.1-8B-Instruct-Turbo'
            model served via TogetherAI
        TOGETHER_LLAMA_3_1_70B_INSTRUCT_TURBO (NDLLMProvider): refers to 'Meta-Llama-3.1-70B-Instruct-Turbo'
            model served via TogetherAI
        TOGETHER_LLAMA_3_1_405B_INSTRUCT_TURBO (NDLLMProvider): refers to 'Meta-Llama-3.1-405B-Instruct-Turbo'
            model served via TogetherAI
        TOGETHER_DEEPSEEK_R1 (NDLLMProvider): refers to 'DeepSeek-R1'
            model served via TogetherAI

        REPLICATE_MISTRAL_7B_INSTRUCT_V0_2 (NDLLMProvider): refers to "mistral-7b-instruct-v0.2" model served via Replicate
        REPLICATE_MIXTRAL_8X7B_INSTRUCT_V0_1 (NDLLMProvider): refers to "mixtral-8x7b-instruct-v0.1" model served via Replicate
        REPLICATE_META_LLAMA_3_70B_INSTRUCT (NDLLMProvider): refers to "meta-llama-3-70b-instruct" model served via Replicate
        REPLICATE_META_LLAMA_3_8B_INSTRUCT (NDLLMProvider): refers to "meta-llama-3-8b-instruct" model served via Replicate
        REPLICATE_META_LLAMA_3_1_405B_INSTRUCT (NDLLMProvider): refers to "meta-llama-3.1-405b-instruct"
            model served via Replicate

        SONAR (NDLLMProvider): refers to "sonar" model by Perplexity
    """

    GPT_3_5_TURBO = ("openai", "gpt-3.5-turbo")
    GPT_3_5_TURBO_0125 = ("openai", "gpt-3.5-turbo-0125")
    GPT_4 = ("openai", "gpt-4")
    GPT_4_0613 = ("openai", "gpt-4-0613")
    GPT_4_1106_PREVIEW = ("openai", "gpt-4-1106-preview")
    GPT_4_TURBO = ("openai", "gpt-4-turbo")
    GPT_4_TURBO_PREVIEW = ("openai", "gpt-4-turbo-preview")
    GPT_4_TURBO_2024_04_09 = ("openai", "gpt-4-turbo-2024-04-09")
    GPT_4o_2024_05_13 = ("openai", "gpt-4o-2024-05-13")
    GPT_4o_2024_08_06 = ("openai", "gpt-4o-2024-08-06")
    GPT_4o = ("openai", "gpt-4o")
    GPT_4o_MINI_2024_07_18 = ("openai", "gpt-4o-mini-2024-07-18")
    GPT_4o_MINI = ("openai", "gpt-4o-mini")
    GPT_4_0125_PREVIEW = ("openai", "gpt-4-0125-preview")
    GPT_4_1 = ("openai", "gpt-4.1")
    GPT_4_1_2025_04_14 = ("openai", "gpt-4.1-2025-04-14")
    GPT_4_1_MINI = ("openai", "gpt-4.1-mini")
    GPT_4_1_MINI_2025_04_14 = ("openai", "gpt-4.1-mini-2025-04-14")
    GPT_4_1_NANO = ("openai", "gpt-4.1-nano")
    GPT_4_1_NANO_2025_04_14 = ("openai", "gpt-4.1-nano-2025-04-14")
    CHATGPT_4o_LATEST = ("openai", "chatgpt-4o-latest")
    GPT_4o_2024_11_20 = ("openai", "gpt-4o-2024-11-20")

    CLAUDE_2_1 = ("anthropic", "claude-2.1")
    CLAUDE_3_OPUS_20240229 = ("anthropic", "claude-3-opus-20240229")
    CLAUDE_3_SONNET_20240229 = ("anthropic", "claude-3-sonnet-20240229")
    CLAUDE_3_5_SONNET_20240620 = ("anthropic", "claude-3-5-sonnet-20240620")
    CLAUDE_3_5_SONNET_20241022 = ("anthropic", "claude-3-5-sonnet-20241022")
    CLAUDE_3_5_SONNET_LATEST = ("anthropic", "claude-3-5-sonnet-latest")
    CLAUDE_3_7_SONNET_LATEST = ("anthropic", "claude-3-7-sonnet-latest")
    CLAUDE_3_7_SONNET_20250219 = ("anthropic", "claude-3-7-sonnet-20250219")
    CLAUDE_3_5_HAIKU_20241022 = ("anthropic", "claude-3-5-haiku-20241022")
    CLAUDE_3_HAIKU_20240307 = ("anthropic", "claude-3-haiku-20240307")
    CLAUDE_OPUS_4_20250514 = ("anthropic", "claude-opus-4-20250514")
    CLAUDE_OPUS_4_1_20250805 = ("anthropic", "claude-opus-4-1-20250805")
    CLAUDE_SONNET_4_20250514 = ("anthropic", "claude-sonnet-4-20250514")
    CLAUDE_SONNET_4_5 = ("anthropic", "claude-sonnet-4-5")
    CLAUDE_SONNET_4_5_20250929 = ("anthropic", "claude-sonnet-4-5-20250929")
    CLAUDE_OPUS_4_0 = ("anthropic", "claude-opus-4-0")
    CLAUDE_OPUS_4_1 = ("anthropic", "claude-opus-4-1")
    CLAUDE_SONNET_4_0 = ("anthropic", "claude-sonnet-4-0")

    GEMINI_PRO = ("google", "gemini-pro")
    GEMINI_1_PRO_LATEST = ("google", "gemini-1.0-pro-latest")
    GEMINI_15_PRO_LATEST = ("google", "gemini-1.5-pro-latest")
    GEMINI_15_PRO_EXP_0801 = ("google", "gemini-1.5-pro-exp-0801")
    GEMINI_15_FLASH_LATEST = ("google", "gemini-1.5-flash-latest")
    GEMINI_20_FLASH = ("google", "gemini-2.0-flash")
    GEMINI_20_FLASH_001 = ("google", "gemini-2.0-flash-001")
    GEMINI_25_FLASH = ("google", "gemini-2.5-flash")
    GEMINI_25_PRO = ("google", "gemini-2.5-pro")

    COMMAND_R = ("cohere", "command-r")
    COMMAND_R_PLUS = ("cohere", "command-r-plus")

    MISTRAL_LARGE_LATEST = ("mistral", "mistral-large-latest")
    MISTRAL_LARGE_2407 = ("mistral", "mistral-large-2407")
    MISTRAL_LARGE_2402 = ("mistral", "mistral-large-2402")
    MISTRAL_MEDIUM_LATEST = ("mistral", "mistral-medium-latest")
    MISTRAL_SMALL_LATEST = ("mistral", "mistral-small-latest")
    CODESTRAL_LATEST = ("mistral", "codestral-latest")
    OPEN_MISTRAL_7B = ("mistral", "open-mistral-7b")
    OPEN_MIXTRAL_8X7B = ("mistral", "open-mixtral-8x7b")
    OPEN_MIXTRAL_8X22B = ("mistral", "open-mixtral-8x22b")
    OPEN_MISTRAL_NEMO = ("mistral", "open-mistral-nemo")

    TOGETHER_MISTRAL_7B_INSTRUCT_V0_2 = (
        "togetherai",
        "Mistral-7B-Instruct-v0.2",
    )
    TOGETHER_MIXTRAL_8X7B_INSTRUCT_V0_1 = (
        "togetherai",
        "Mixtral-8x7B-Instruct-v0.1",
    )
    TOGETHER_MIXTRAL_8X22B_INSTRUCT_V0_1 = (
        "togetherai",
        "Mixtral-8x22B-Instruct-v0.1",
    )
    TOGETHER_LLAMA_3_70B_CHAT_HF = ("togetherai", "Llama-3-70b-chat-hf")
    TOGETHER_LLAMA_3_8B_CHAT_HF = ("togetherai", "Llama-3-8b-chat-hf")
    TOGETHER_QWEN2_72B_INSTRUCT = ("togetherai", "Qwen2-72B-Instruct")
    TOGETHER_LLAMA_3_1_8B_INSTRUCT_TURBO = (
        "togetherai",
        "Meta-Llama-3.1-8B-Instruct-Turbo",
    )
    TOGETHER_LLAMA_3_1_70B_INSTRUCT_TURBO = (
        "togetherai",
        "Meta-Llama-3.1-70B-Instruct-Turbo",
    )
    TOGETHER_LLAMA_3_1_405B_INSTRUCT_TURBO = (
        "togetherai",
        "Meta-Llama-3.1-405B-Instruct-Turbo",
    )
    TOGETHER_DEEPSEEK_R1 = ("togetherai", "DeepSeek-R1")

    SONAR = (
        "perplexity",
        "sonar",
    )

    REPLICATE_MISTRAL_7B_INSTRUCT_V0_2 = (
        "replicate",
        "mistral-7b-instruct-v0.2",
    )
    REPLICATE_MIXTRAL_8X7B_INSTRUCT_V0_1 = (
        "replicate",
        "mixtral-8x7b-instruct-v0.1",
    )
    REPLICATE_META_LLAMA_3_70B_INSTRUCT = (
        "replicate",
        "meta-llama-3-70b-instruct",
    )
    REPLICATE_META_LLAMA_3_8B_INSTRUCT = (
        "replicate",
        "meta-llama-3-8b-instruct",
    )
    REPLICATE_META_LLAMA_3_1_405B_INSTRUCT = (
        "replicate",
        "meta-llama-3.1-405b-instruct",
    )

    def __new__(cls, provider, model):
        return LLMConfig(provider=provider, model=model)





import logging
import re
from typing import Dict, List

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def inject_system_prompt(
    messages: List[Dict[str, str]], system_prompt: str
) -> List[Dict[str, str]]:
    """
    Add a system prompt to an OpenAI-style message list. If a system prompt is already present, replace it.
    """
    new_messages = []
    found = False
    for msg in messages:
        # t7: replace the first system prompt with the new one
        if msg["role"] == "system" and not found:
            new_messages.append({"role": "system", "content": system_prompt})
            found = True
        else:
            new_messages.append(msg)
    if not found:
        new_messages.insert(0, {"role": "system", "content": system_prompt})
    return new_messages


def _curly_escape(text: str) -> str:
    """
    Escape curly braces in the text, but only for single occurrences of alphabetic characters.
    This function will not escape double curly braces or non-alphabetic characters.
    """
    return re.sub(r"(?<!{){([a-zA-Z])}(?!})", r"{{\1}}", text)





import os
from importlib import metadata
from typing import (
    Any,
    AsyncIterator,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Union,
)

from langchain.chat_models.base import init_chat_model
from langchain_community.adapters.openai import convert_message_to_dict
from langchain_core.language_models import LanguageModelInput
from langchain_core.messages.utils import convert_to_messages
from langchain_core.prompt_values import (
    ChatPromptValue,
    PromptValue,
    StringPromptValue,
)
from langchain_core.runnables import Runnable, RunnableConfig

import notdiamond as nd

_LANGCHAIN_PROVIDERS = {
    "openai",
    "anthropic",
    "google",
    "mistral",
    "togetherai",
    "cohere",
}


class NotDiamondRunnable(Runnable[LanguageModelInput, str]):
    """
    See Runnable docs for details
    https://python.langchain.com/v0.1/docs/expression_language/interface/
    """

    llm_configs: List
    api_key: Optional[str] = os.getenv("NOTDIAMOND_API_KEY")
    client: Any

    def __init__(
        self,
        nd_llm_configs: Optional[List] = None,
        nd_api_key: Optional[str] = None,
        nd_client: Optional[Any] = None,
        nd_kwargs: Optional[Dict[str, Any]] = None,
    ):
        """
        Params:
            nd_llm_configs: List of LLM configs to use.
            nd_api_key: Not Diamond API key.
            nd_client: Not Diamond client.
            nd_kwargs: Keyword arguments to pass directly to model_select.
        """
        if not nd_client:
            if not nd_api_key or not nd_llm_configs:
                raise ValueError(
                    "Must provide either client or api_key and llm_configs to "
                    "instantiate NotDiamondRunnable."
                )
            nd_client = nd.NotDiamond(
                llm_configs=nd_llm_configs,
                api_key=nd_api_key,
            )
        elif nd_client.llm_configs:
            for llm_config in nd_client.llm_configs:
                if isinstance(llm_config, str):
                    llm_config = nd.LLMConfig.from_string(llm_config)
                if llm_config.provider not in _LANGCHAIN_PROVIDERS:
                    raise ValueError(
                        f"Requested provider in {llm_config} supported by Not Diamond "
                        "but not langchain.chat_models.base.init_chat_model. Please "
                        "remove it from your llm_configs."
                    )

        try:
            nd_client.user_agent = (
                f"langchain-community/{metadata.version('notdiamond')}"
            )
        except AttributeError:
            pass

        self.client = nd_client
        self.api_key = nd_client.api_key
        self.llm_configs = nd_client.llm_configs
        self.nd_kwargs = nd_kwargs or dict()

    def _model_select(self, input: LanguageModelInput) -> str:
        messages = _convert_input_to_message_dicts(input)
        _, provider = self.client.chat.completions.model_select(
            messages=messages, **self.nd_kwargs
        )
        provider_str = _nd_provider_to_langchain_provider(str(provider))
        return provider_str

    async def _amodel_select(self, input: LanguageModelInput) -> str:
        messages = _convert_input_to_message_dicts(input)
        _, provider = await self.client.chat.completions.amodel_select(
            messages=messages, **self.nd_kwargs
        )
        provider_str = _nd_provider_to_langchain_provider(str(provider))
        return provider_str

    def stream(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Iterator[str]:
        yield self._model_select(input)

    def invoke(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
    ) -> str:
        return self._model_select(input)

    def batch(
        self,
        inputs: List[LanguageModelInput],
        config: Optional[Union[RunnableConfig, List[RunnableConfig]]] = None,
        **kwargs: Optional[Any],
    ) -> List[str]:
        return [self._model_select(input) for input in inputs]

    async def astream(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> AsyncIterator[str]:
        yield await self._amodel_select(input)

    async def ainvoke(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> str:
        return await self._amodel_select(input)

    async def abatch(
        self,
        inputs: List[LanguageModelInput],
        config: Optional[Union[RunnableConfig, List[RunnableConfig]]] = None,
        **kwargs: Optional[Any],
    ) -> List[str]:
        return [await self._amodel_select(input) for input in inputs]


class NotDiamondRoutedRunnable(Runnable[LanguageModelInput, Any]):
    def __init__(
        self,
        *args: Any,
        configurable_fields: Optional[List[str]] = None,
        nd_llm_configs: Optional[List] = None,
        nd_api_key: Optional[str] = None,
        nd_client: Optional[Any] = None,
        nd_kwargs: Optional[Dict[str, Any]] = None,
        **kwargs: Optional[Dict[Any, Any]],
    ) -> None:
        """
        Params:
            nd_llm_configs: List of LLM configs to use.
            nd_api_key: Not Diamond API key.
            nd_client: Not Diamond client.
            nd_kwargs: Keyword arguments to pass directly to model_select.
        """
        _nd_kwargs = {
            kw: kwargs[kw] for kw in kwargs.keys() if kw.startswith("nd_")
        }
        if nd_kwargs:
            _nd_kwargs.update(nd_kwargs)

        self._ndrunnable = NotDiamondRunnable(
            nd_api_key=nd_api_key,
            nd_llm_configs=nd_llm_configs,
            nd_client=nd_client,
            nd_kwargs=_nd_kwargs,
        )
        _routed_fields = ["model", "model_provider"]
        if configurable_fields is None:
            configurable_fields = []
        self._configurable_fields = _routed_fields + configurable_fields
        self._configurable_model = init_chat_model(
            *args,
            configurable_fields=self._configurable_fields,
            config_prefix="nd",
            **{kw: kwv for kw, kwv in kwargs.items() if kw not in _nd_kwargs},  # type: ignore[arg-type]
        )

    def stream(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Iterator[Any]:
        provider_str = self._ndrunnable._model_select(input)
        _config = self._build_model_config(provider_str, config)
        yield from self._configurable_model.stream(input, config=_config)

    def invoke(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Any:
        provider_str = self._ndrunnable._model_select(input)
        _config = self._build_model_config(provider_str, config)
        return self._configurable_model.invoke(input, config=_config)

    def batch(
        self,
        inputs: List[LanguageModelInput],
        config: Optional[Union[RunnableConfig, List[RunnableConfig]]] = None,
        **kwargs: Optional[Any],
    ) -> List[Any]:
        config = config or {}

        provider_strs = [
            self._ndrunnable._model_select(input) for input in inputs
        ]
        if isinstance(config, dict):
            _configs = [
                self._build_model_config(ps, config) for ps in provider_strs
            ]
        else:
            _configs = [
                self._build_model_config(ps, config[i])
                for i, ps in enumerate(provider_strs)
            ]

        return self._configurable_model.batch(inputs, config=_configs)

    async def astream(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> AsyncIterator[Any]:
        provider_str = await self._ndrunnable._amodel_select(input)
        _config = self._build_model_config(provider_str, config)
        async for chunk in self._configurable_model.astream(
            input, config=_config
        ):
            yield chunk

    async def ainvoke(
        self,
        input: LanguageModelInput,
        config: Optional[RunnableConfig] = None,
        **kwargs: Optional[Any],
    ) -> Any:
        provider_str = await self._ndrunnable._amodel_select(input)
        _config = self._build_model_config(provider_str, config)
        return await self._configurable_model.ainvoke(input, config=_config)

    async def abatch(
        self,
        inputs: List[LanguageModelInput],
        config: Optional[Union[RunnableConfig, List[RunnableConfig]]] = None,
        **kwargs: Optional[Any],
    ) -> List[Any]:
        config = config or {}

        provider_strs = [
            await self._ndrunnable._amodel_select(input) for input in inputs
        ]
        if isinstance(config, dict):
            _configs = [
                self._build_model_config(ps, config) for ps in provider_strs
            ]
        else:
            _configs = [
                self._build_model_config(ps, config[i])
                for i, ps in enumerate(provider_strs)
            ]

        return await self._configurable_model.abatch(inputs, config=_configs)

    def _build_model_config(
        self, provider_str: str, config: Optional[RunnableConfig] = None
    ) -> RunnableConfig:
        """
        Provider string should take the form '{model}/{model_provider}'
        """
        config = config or RunnableConfig()

        model_provider, model = provider_str.split("/")
        _config = RunnableConfig(
            configurable={
                "nd_model": model,
                "nd_model_provider": model_provider,
            },
        )

        for k, v in config.items():
            _config["configurable"][f"nd_{k}"] = v
        return _config


def _convert_input_to_message_dicts(
    input: LanguageModelInput,
) -> List[Dict[str, str]]:
    if isinstance(input, PromptValue):
        output = input
    elif isinstance(input, str):
        output = StringPromptValue(text=input)
    elif isinstance(input, Sequence):
        output = ChatPromptValue(messages=convert_to_messages(input))
    else:
        raise ValueError(
            f"Invalid input type {type(input)}. "
            "Must be a PromptValue, str, or list of BaseMessages."
        )
    return [
        convert_message_to_dict(message) for message in output.to_messages()
    ]


def _nd_provider_to_langchain_provider(llm_config_str: str) -> str:
    provider, model = llm_config_str.split("/")
    provider = (
        provider.replace("google", "google_genai")
        .replace("mistral", "mistralai")
        .replace("togetherai", "together")
    )
    return f"{provider}/{model}"



from notdiamond.llms.client import NotDiamond  # noqa: F401
from notdiamond.llms.config import LLMConfig  # noqa: F401
from notdiamond.metrics.metric import Metric  # noqa: F401

__all__ = []
try:
    import importlib.util

    spec = importlib.util.find_spec("openai")
    if spec is not None:
        from ._init import init  # noqa: F401

        __all__ = ["init"]
    else:
        __all__ = []
except ImportError:
    pass



from typing import List, Optional, Sequence, Union

import pandas as pd
from langchain_core.callbacks import Callbacks
from langchain_core.documents import Document as LCDocument
from llama_index.core.base.embeddings.base import (
    BaseEmbedding as LlamaIndexEmbedding,
)
from llama_index.core.base.llms.base import BaseLLM as LlamaIndexLLM
from llama_index.core.schema import Document as LlamaIndexDocument
from ragas.embeddings import BaseRagasEmbeddings, LlamaIndexEmbeddingsWrapper
from ragas.llms import BaseRagasLLM, LlamaIndexLLMWrapper
from ragas.run_config import RunConfig
from ragas.testset import TestsetGenerator
from ragas.testset.graph import KnowledgeGraph, Node, NodeType
from ragas.testset.persona import Persona
from ragas.testset.synthesizers import QueryDistribution
from ragas.testset.transforms import (
    Transforms,
    apply_transforms,
    default_transforms,
)


class TestDataGenerator(TestsetGenerator):
    def __init__(
        self,
        llm: BaseRagasLLM,
        embedding_model: BaseRagasEmbeddings,
        knowledge_graph: KnowledgeGraph = KnowledgeGraph(),
        persona_list: Optional[List[Persona]] = None,
    ):
        """
        RAG Test data generator class.
        Generates test cases from documents for evaluating RAG workflows.

        Parameters:
            llm (BaseRagasLLM): An LLM object inherited from BaseRagasLLM. Obtain this
                via the get_llm tool.
            embedding_model (BaseRagasEmbeddings): An embedding model object inherited
                from BaseRagasEmbeddings. Obtain this via the get_embedding tool.
            knowledge_graph (KnowledgeGraph): The knowledge graph to use for the generation
                process. Default empty.
        """
        super().__init__(
            llm=llm,
            embedding_model=embedding_model,
            knowledge_graph=knowledge_graph,
            persona_list=persona_list,
        )

    def generate_from_docs(
        self,
        documents: Union[Sequence[LCDocument], Sequence[LlamaIndexDocument]],
        testset_size: int,
        transforms: Optional[Transforms] = None,
        transforms_llm: Optional[Union[BaseRagasLLM, LlamaIndexLLM]] = None,
        transforms_embedding_model: Optional[
            Union[BaseRagasEmbeddings, LlamaIndexEmbedding]
        ] = None,
        query_distribution: Optional[QueryDistribution] = None,
        run_config: Optional[RunConfig] = None,
        callbacks: Optional[Callbacks] = None,
        with_debugging_logs: bool = False,
        raise_exceptions: bool = True,
    ) -> pd.DataFrame:
        """
        Generates an evaluation dataset based on given Langchain or Llama Index documents and parameters.

        Parameters:
            documents : Sequence[LCDocument]
                A sequence of Langchain documents to use as source material
            testset_size : int
                The number of test samples to generate
            transforms : Optional[Transforms], optional
                Custom transforms to apply to the documents, by default None
            transforms_llm : Optional[BaseRagasLLM], optional
                LLM to use for transforms if different from instance LLM, by default None
            transforms_embedding_model : Optional[BaseRagasEmbeddings], optional
                Embedding model to use for transforms if different from instance model, by default None
            query_distribution : Optional[QueryDistribution], optional
                Distribution of query types to generate, by default None
            run_config : Optional[RunConfig], optional
                Configuration for the generation run, by default None
            callbacks : Optional[Callbacks], optional
                Callbacks to use during generation, by default None
            with_debugging_logs : bool, optional
                Whether to include debug logs, by default False
            raise_exceptions : bool, optional
                Whether to raise exceptions during generation, by default True

        Returns:
            Testset
                The generated evaluation dataset

        Raises:
            ValueError
                If no LLM or embedding model is provided either during initialization or as arguments
        """
        assert isinstance(
            documents, list
        ), "Documents must be a list of langchain or llama-index documents."

        if isinstance(documents[0], LCDocument):
            dataset = self.generate_with_langchain_docs(
                documents=documents,
                testset_size=testset_size,
                transforms=transforms,
                transforms_llm=transforms_llm,
                transforms_embedding_model=transforms_embedding_model,
                query_distribution=query_distribution,
                run_config=run_config,
                callbacks=callbacks,
                with_debugging_logs=with_debugging_logs,
                raise_exceptions=raise_exceptions,
            )
            return dataset.to_pandas()

        elif isinstance(documents[0], LlamaIndexDocument):
            dataset = self.generate_with_llamaindex_docs(
                documents=documents,
                testset_size=testset_size,
                transforms=transforms,
                transforms_llm=transforms_llm,
                transforms_embedding_model=transforms_embedding_model,
                query_distribution=query_distribution,
                run_config=run_config,
                callbacks=callbacks,
                with_debugging_logs=with_debugging_logs,
                raise_exceptions=raise_exceptions,
            )
            return dataset.to_pandas()

        raise ValueError(
            "Documents must be a list of langchain or llama-index documents."
        )

    def generate_with_llamaindex_docs(
        self,
        documents: Sequence[LlamaIndexDocument],
        testset_size: int,
        transforms: Optional[Transforms] = None,
        transforms_llm: Optional[LlamaIndexLLM] = None,
        transforms_embedding_model: Optional[LlamaIndexEmbedding] = None,
        query_distribution: Optional[QueryDistribution] = None,
        run_config: Optional[RunConfig] = None,
        callbacks: Optional[Callbacks] = None,
        with_debugging_logs=False,
        raise_exceptions: bool = True,
    ):
        """
        Generates an evaluation dataset based on given scenarios and parameters.
        """

        run_config = run_config or RunConfig()

        # force the user to provide an llm and embedding client to prevent use of default LLMs
        if not self.llm and not transforms_llm:
            raise ValueError(
                "An llm client was not provided."
                " Provide an LLM on init or as an argument to this method."
                " Alternatively you can provide your own transforms through the `transforms` parameter."
            )
        if not self.embedding_model and not transforms_embedding_model:
            raise ValueError(
                "An embedding client was not provided."
                " Provide an embedding through the transforms_embedding_model parameter."
                " Alternatively you can provide your own transforms through the `transforms` parameter."
            )

        if not transforms:
            # use TestsetGenerator's LLM and embedding model if no transforms_llm or transforms_embedding_model is provided
            if transforms_llm is None:
                llm_for_transforms = self.llm
            else:
                llm_for_transforms = LlamaIndexLLMWrapper(transforms_llm)
            if transforms_embedding_model is None:
                embedding_model_for_transforms = self.embedding_model
            else:
                embedding_model_for_transforms = LlamaIndexEmbeddingsWrapper(
                    transforms_embedding_model
                )

            # create the transforms
            transforms = default_transforms(
                documents=[
                    LCDocument(page_content=doc.text) for doc in documents
                ],
                llm=llm_for_transforms,
                embedding_model=embedding_model_for_transforms,
            )

        # convert the documents to Ragas nodes
        nodes = []
        for doc in documents:
            if doc.text is not None and doc.text.strip() != "":
                node = Node(
                    type=NodeType.DOCUMENT,
                    properties={
                        "page_content": doc.text,
                        "document_metadata": doc.metadata,
                    },
                )
                nodes.append(node)

        kg = KnowledgeGraph(nodes=nodes)

        # apply transforms and update the knowledge graph
        apply_transforms(kg, transforms, run_config)
        self.knowledge_graph = kg

        return self.generate(
            testset_size=testset_size,
            query_distribution=query_distribution,
            run_config=run_config,
            callbacks=callbacks,
            with_debugging_logs=with_debugging_logs,
            raise_exceptions=raise_exceptions,
        )



import json
import logging
from typing import Any, Callable, Dict, List, Optional, Sequence, Union

import aiohttp
import requests

from notdiamond import settings
from notdiamond._utils import _default_headers, convert_tool_to_openai_function
from notdiamond.llms.config import LLMConfig
from notdiamond.metrics.metric import Metric
from notdiamond.types import ModelSelectRequestPayload

LOGGER = logging.getLogger(__name__)
LOGGER.setLevel(logging.INFO)


def model_select_prepare(
    messages: List[Dict[str, str]],
    llm_configs: List[LLMConfig],
    metric: Metric,
    notdiamond_api_key: str,
    max_model_depth: int,
    hash_content: bool,
    tradeoff: Optional[str] = None,
    preference_id: Optional[str] = None,
    tools: Optional[Sequence[Union[Dict[str, Any], Callable]]] = [],
    previous_session: Optional[str] = None,
    nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
    _user_agent: str = settings.DEFAULT_USER_AGENT,
):
    """
    This is the core method for the model_select endpoint.
    It returns the best fitting LLM to call and a session ID that can be used for feedback.

    Parameters:
        messages (List[Dict[str, str]]): list of messages to be used for the LLM call
        llm_configs (List[LLMConfig]): a list of available LLMs that the router can decide from
        metric (Metric): metric based off which the router makes the decision. As of now only 'accuracy' supported.
        notdiamond_api_key (str): API key generated via the NotDiamond dashboard.
        max_model_depth (int): if your top recommended model is down, specify up to which depth of routing you're willing to go.
        hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
        tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                            for the router to determine the best LLM for a given query.
        preference_id (Optional[str], optional): The ID of the router preference that was configured via the Dashboard.
                                                    Defaults to None.
        previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
        async_mode (bool, optional): whether to run the request in async mode. Defaults to False.
        nd_api_url (Optional[str], optional): The URL of the NotDiamond API. Defaults to None.

    Returns:
        tuple(url, payload, headers): returns data to be used for the API call of modelSelect
    """
    url = f"{nd_api_url}/v2/modelRouter/modelSelect"
    tools_dict = get_tools_in_openai_format(tools)

    payload: ModelSelectRequestPayload = {
        "messages": messages,
        "llm_providers": [
            llm_provider.prepare_for_request() for llm_provider in llm_configs
        ],
        "metric": metric.metric,
        "max_model_depth": max_model_depth,
        "hash_content": hash_content,
    }

    if tools_dict:
        payload["tools"] = tools_dict
    if tradeoff is not None:
        payload["tradeoff"] = tradeoff
    if preference_id is not None:
        payload["preference_id"] = preference_id
    if previous_session is not None:
        payload["previous_session"] = previous_session

    headers = _default_headers(notdiamond_api_key, _user_agent)

    return url, payload, headers


def get_tools_in_openai_format(
    tools: Optional[Sequence[Union[Dict[str, Any], Callable]]],
):
    """
    This function converts the tools list into the format that OpenAI expects.
    Does this by using langchains Model that automatically creates the dictionary on bind_tools

    Parameters:
        tools (Optional[Sequence[Union[Dict[str, Any], Callable]]]): list of tools to be converted

    Returns:
        dict: dictionary of tools in the format that OpenAI expects
    """
    if tools:
        return [
            {
                "type": "function",
                "function": convert_tool_to_openai_function(tool),
            }
            for tool in tools
        ]

    return None


def model_select_parse(response_code, response_json, llm_configs):
    if response_code == 200:
        providers = response_json["providers"]
        session_id = response_json["session_id"]

        top_provider = providers[0]

        best_llm = list(
            filter(
                lambda x: (x.model == top_provider["model"])
                & (x.provider == top_provider["provider"]),
                llm_configs,
            )
        )[0]
        return best_llm, session_id

    error_message = response_json["detail"]
    LOGGER.error(f"API error: {response_code}. {error_message}")
    return None, "NO-SESSION-ID"


def model_select(
    messages: List[Dict[str, str]],
    llm_configs: List[LLMConfig],
    metric: Metric,
    notdiamond_api_key: str,
    max_model_depth: int,
    hash_content: bool,
    tradeoff: Optional[str] = None,
    preference_id: Optional[str] = None,
    tools: Optional[Sequence[Union[Dict[str, Any], Callable]]] = [],
    previous_session: Optional[str] = None,
    timeout: Optional[Union[float, int]] = 60,
    max_retries: Optional[int] = 3,
    nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
    _user_agent: str = settings.DEFAULT_USER_AGENT,
):
    """
    This endpoint receives the prompt and routing settings, and makes a call to the NotDiamond API.
    It returns the best fitting LLM to call and a session ID that can be used for feedback.

    Parameters:
        messages (List[Dict[str, str]]): list of messages to be used for the LLM call
        llm_configs (List[LLMConfig]): a list of available LLMs that the router can decide from
        metric (Metric): metric based off which the router makes the decision. As of now only 'accuracy' supported.
        notdiamond_api_key (str): API key generated via the NotDiamond dashboard.
        max_model_depth (int): if your top recommended model is down, specify up to which depth of routing you're willing to go.
        hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
        tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                            for the router to determine the best LLM for a given query.
        preference_id (Optional[str], optional): The ID of the router preference that was configured via the Dashboard.
                                                    Defaults to None.
        previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
        timeout (int, optional): timeout for the request. Defaults to 60.
        max_retries (int, optional): The maximum number of retries to make when calling the Not Diamond API.
                                    Defaults to 3.
        nd_api_url (Optional[str], optional): The URL of the NotDiamond API. Defaults to None.
    Returns:
        tuple(LLMConfig, string): returns a tuple of the chosen LLMConfig to call and a session ID string.
                                        In case of an error the LLM defaults to None and the session ID defaults
                                        to 'NO-SESSION-ID'.
    """
    url, payload, headers = model_select_prepare(
        messages=messages,
        llm_configs=llm_configs,
        metric=metric,
        notdiamond_api_key=notdiamond_api_key,
        max_model_depth=max_model_depth,
        hash_content=hash_content,
        tradeoff=tradeoff,
        preference_id=preference_id,
        tools=tools,
        previous_session=previous_session,
        nd_api_url=nd_api_url,
        _user_agent=_user_agent,
    )

    for n_retry in range(1, max_retries + 1):
        try:
            response = requests.post(
                url, data=json.dumps(payload), headers=headers, timeout=timeout
            )
            response_code = response.status_code
            response_json = response.json()
            break
        except Exception as e:
            LOGGER.error(
                f"Retry {n_retry} of {max_retries}: API error: {e}",
                exc_info=True,
            )
            if n_retry == max_retries:
                return None, "NO-SESSION-ID"

    best_llm, session_id = model_select_parse(
        response_code, response_json, llm_configs
    )

    return best_llm, session_id


async def amodel_select(
    messages: List[Dict[str, str]],
    llm_configs: List[LLMConfig],
    metric: Metric,
    notdiamond_api_key: str,
    max_model_depth: int,
    hash_content: bool,
    tradeoff: Optional[str] = None,
    preference_id: Optional[str] = None,
    tools: Optional[Sequence[Union[Dict[str, Any], Callable]]] = [],
    previous_session: Optional[str] = None,
    timeout: Optional[Union[float, int]] = 60,
    max_retries: Optional[int] = 3,
    nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
    _user_agent: str = settings.DEFAULT_USER_AGENT,
):
    """
    This endpoint receives the prompt and routing settings, and makes a call to the NotDiamond API.
    It returns the best fitting LLM to call and a session ID that can be used for feedback.

    Parameters:
        messages (List[Dict[str, str]]): list of messages to be used for the LLM call
        llm_configs (List[LLMConfig]): a list of available LLMs that the router can decide from
        metric (Metric): metric based off which the router makes the decision. As of now only 'accuracy' supported.
        notdiamond_api_key (str): API key generated via the NotDiamond dashboard.
        max_model_depth (int): if your top recommended model is down, specify up to which depth of routing you're willing to go.
        hash_content (Optional[bool]): Flag for hashing content before sending to NotDiamond API.
        tradeoff (Optional[str], optional): Define the "cost" or "latency" tradeoff
                                            for the router to determine the best LLM for a given query.
        preference_id (Optional[str], optional): The ID of the router preference that was configured via the Dashboard.
                                                    Defaults to None.
        previous_session (Optional[str], optional): The session ID of a previous session, allow you to link requests.
        timeout (int, optional): timeout for the request. Defaults to 60.
        max_retries (int, optional): The maximum number of retries to make when calling the Not Diamond API.
        nd_api_url (Optional[str], optional): The URL of the NotDiamond API. Defaults to None.
    Returns:
        tuple(LLMConfig, string): returns a tuple of the chosen LLMConfig to call and a session ID string.
                                        In case of an error the LLM defaults to None and the session ID defaults
                                        to 'NO-SESSION-ID'.
    """
    url, payload, headers = model_select_prepare(
        messages=messages,
        llm_configs=llm_configs,
        metric=metric,
        notdiamond_api_key=notdiamond_api_key,
        max_model_depth=max_model_depth,
        hash_content=hash_content,
        tradeoff=tradeoff,
        preference_id=preference_id,
        tools=tools,
        previous_session=previous_session,
        nd_api_url=nd_api_url,
        _user_agent=_user_agent,
    )

    for n_retry in range(1, max_retries + 1):
        try:
            async with aiohttp.ClientSession() as session:
                async with session.post(
                    url,
                    data=json.dumps(payload),
                    headers=headers,
                    timeout=timeout,
                ) as response:
                    response_code = response.status
                    response_json = await response.json()
                    break
        except Exception as e:
            LOGGER.error(
                f"Retry {n_retry} of {max_retries}: API error: {e}",
                exc_info=True,
            )
            if n_retry == max_retries:
                return None, "NO-SESSION-ID"

    best_llm, session_id = model_select_parse(
        response_code, response_json, llm_configs
    )

    return best_llm, session_id


def report_latency(
    session_id: str,
    llm_config: LLMConfig,
    tokens_per_second: float,
    notdiamond_api_key: str,
    nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
    _user_agent: str = settings.DEFAULT_USER_AGENT,
):
    """
    This method makes an API call to the NotDiamond server to report the latency of an LLM call.
    It helps fine-tune our model router and ensure we offer recommendations that meet your latency expectation.

    This feature can be disabled on the NDLLM class level by setting `latency_tracking` to False.

    Parameters:
        session_id (str): the session ID that was returned from the `invoke` or `model_select` calls, so we know which
                            router call your latency report refers to.
        llm_provider (LLMConfig): specifying the LLM provider for which the latency is reported
        tokens_per_second (float): latency of the model call calculated based on time elapsed, input tokens, and output tokens
        notdiamond_api_key (str): NotDiamond API call used for authentication
        nd_api_url (Optional[str], optional): The URL of the NotDiamond API. Defaults to None.
    Returns:
        int: status code of the API call, 200 if it's success

    Raises:
        ApiError: if the API call to the NotDiamond backend fails, this error is raised
    """
    url = f"{nd_api_url}/v2/report/metrics/latency"

    payload = {
        "session_id": session_id,
        "provider": llm_config.prepare_for_request(),
        "feedback": {"tokens_per_second": tokens_per_second},
    }

    headers = _default_headers(notdiamond_api_key, _user_agent)

    try:
        response = requests.post(url, json=payload, headers=headers)
    except Exception as e:
        LOGGER.error(
            f"API error for report metrics latency: {e}", exc_info=True
        )
        return 500

    return response.status_code


def create_preference_id(
    notdiamond_api_key: str,
    name: Optional[str] = None,
    nd_api_url: Optional[str] = settings.NOTDIAMOND_API_URL,
    _user_agent: str = settings.DEFAULT_USER_AGENT,
) -> str:
    """
    Create a preference id with an optional name. The preference name will appear in your
    dashboard on Not Diamond.
    """
    url = f"{nd_api_url}/v2/preferences/userPreferenceCreate"
    headers = _default_headers(notdiamond_api_key, _user_agent)
    res = requests.post(url=url, headers=headers, json={"name": name})
    if res.status_code == 200:
        preference_id = res.json()["preference_id"]
    else:
        raise Exception(f"Error creating preference ID: {res.text}")

    return preference_id
